# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1.exceptions import base

__all__ = ('ServiceProviderNotFound',)


class ServiceProviderNotFound(base.ClientException):
    """A Service Provider cannot be found."""

    def __init__(self, sp_id: str):
        self.sp_id = sp_id
        msg = f'The Service Provider {sp_id} could not be found'
        super().__init__(msg)
